# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

from ..core.datetime_utils import serialize_datetime

try:
    import pydantic.v1 as pydantic  # type: ignore
except ImportError:
    import pydantic  # type: ignore


class Span(pydantic.BaseModel):
    parent_span_id: typing.Optional[str]
    meta: typing.Optional[typing.Dict[str, typing.Any]]
    event_name: str
    event_type: typing.Optional[str]
    start_time: dt.datetime
    duration: typing.Optional[int]
    status: str
    end_time: dt.datetime
    inputs: typing.Optional[typing.List[str]]
    outputs: typing.Optional[typing.List[str]]
    prompt_template: typing.Optional[str]
    tokens_input: typing.Optional[int]
    tokens_output: typing.Optional[int]
    token_total: typing.Optional[int]
    cost: typing.Optional[float]
    tags: typing.Optional[typing.List[str]]
    span_id: str

    def json(self, **kwargs: typing.Any) -> str:
        kwargs_with_defaults: typing.Any = {
            "by_alias": True,
            "exclude_unset": True,
            **kwargs,
        }
        return super().json(**kwargs_with_defaults)

    def dict(self, **kwargs: typing.Any) -> typing.Dict[str, typing.Any]:
        kwargs_with_defaults: typing.Any = {
            "by_alias": True,
            "exclude_unset": True,
            **kwargs,
        }
        return super().dict(**kwargs_with_defaults)

    class Config:
        frozen = True
        smart_union = True
        json_encoders = {dt.datetime: serialize_datetime}
